#  MIT License Copyright (c) 2020. Houfu Ang

"""
Looks over the PDPC website and creates PDPC Decision objects

Requirements:
* Chrome Webdriver to automate web browser
"""
import logging
from typing import List

from selenium.webdriver import Chrome
from selenium.webdriver.chrome.options import Options

from pdpc_decisions.classes import PDPCDecisionItem

logger = logging.getLogger(__name__)


class Scraper:
    def __init__(self):
        options = Options()
        options.add_argument('--headless')
        options.add_argument('--disable-gpu')
        options.add_argument('--window-size=1920,1080')
        options.add_argument('--no-sandbox')

        self.driver = Chrome(options=options)
        self.driver.implicitly_wait(25)

    @classmethod
    def scrape(cls,
               site_url="https://www.pdpc.gov.sg/All-Commissions-Decisions?"
                        "keyword=&industry=all&nature=all&decision=all&penalty=all&page=1") -> List[PDPCDecisionItem]:
        """Convenience method for scraping the PDPC Decision website with Scraper."""
        logger.info('Starting the scrape')
        self = cls()
        result = []
        try:
            self.driver.get(site_url)
            finished = False
            page_count = 1
            logger.info('Now at page 1.')
            while not finished:
                items = self.driver.find_element_by_class_name('listing__list').find_elements_by_tag_name('li')
                for current_item in range(0, len(items)):
                    items = self.driver.find_element_by_class_name('listing__list').find_elements_by_tag_name('li')
                    link = items[current_item].find_element_by_tag_name('a').get_property('href')
                    item = PDPCDecisionItem.from_web_page(link)
                    logger.info(f'Added: {item.respondent}, {item.published_date}')
                    result.append(item)
                next_page = self.driver.find_element_by_class_name('pagination-next')
                if 'disabled' in next_page.get_attribute('class'):
                    logger.info('Scraper has reached end of page.')
                    finished = True
                else:
                    page_count += 1
                    new_url = "https://www.pdpc.gov.sg/All-Commissions-Decisions?page={}".format(page_count)
                    self.driver.get(new_url)
                    logger.info(f'Now at page {page_count}')
        finally:
            self.driver.close()
        logger.info(f'Ending scrape with {len(result)} items.')
        return result



#  MIT License Copyright (c) 2020. Houfu Ang
import io
import logging
import os
from typing import Iterable

import requests

from pdpc_decisions.classes import Options, PDPCDecisionItem, PDFFile

logger = logging.getLogger(__name__)


def download_files(options: Options, items: Iterable[PDPCDecisionItem]):
    logger.info('Start downloading files.')
    if not os.path.exists(options["download_folder"]):
        os.mkdir(options["download_folder"])
    for item in items:
        logger.info(f"Downloading {item}: {item.download_url}")
        if check_pdf(item.download_url):
            download_pdf(options["download_folder"], item)
        else:
            download_text(options["download_folder"], item)
    logger.info(f'Finished downloading files to {options["download_folder"]}')


def check_pdf(download_url: str) -> bool:
    """Check if the download_url is a PDF or not by reading the extension."""
    from urllib.parse import urlparse
    result = urlparse(download_url)
    return result.path[-3:] == 'pdf'


def download_pdf(download_folder: str, item: PDPCDecisionItem) -> str:
    """Downloads a pdf in the item to the download folder. Returns the path where the file is saved."""
    destination = os.path.join(download_folder,
                               f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent}.pdf")
    if os.path.isfile(os.path.join(download_folder,
                                   f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} (1).pdf")):
        logger.warning('Found secondary format for PDF files. To follow format for this file.')
        file_num = 2
        destination = os.path.join(download_folder,
                                   f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} ({file_num}).pdf")
        while os.path.isfile(destination):
            file_num += 1
            destination = os.path.join(download_folder,
                                       f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} ({file_num}).pdf")
    if os.path.isfile(destination):
        logger.warning('Found two decisions with the same PDF output file name. Rewriting to secondary format.')
        file_num = 1
        os.rename(destination,
                  os.path.join(download_folder,
                               f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} ({file_num}).pdf"))
        file_num += 1
        destination = os.path.join(download_folder,
                                   f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} ({file_num}).pdf")
    with open(destination, 'wb') as file:
        pdf_file = requests.get(item.download_url).content
        file.write(pdf_file)
    logger.info(f"Downloaded a pdf: {destination}")
    return destination


def download_text(download_folder: str, item: PDPCDecisionItem) -> str:
    """Downloads a text file in the item to the download folder. Returns the path where the file is saved."""
    destination_filename = "{} {}.txt".format(item.published_date.strftime('%Y-%m-%d'),
                                              item.respondent)
    destination = os.path.join(download_folder, destination_filename)
    with open(destination, "w", encoding='utf-8') as f:
        f.writelines(get_text_stream(item))
    logger.info(f"Downloaded a text: {destination}")
    return destination


def get_text_from_pdf(source: PDPCDecisionItem, options: Options = None) -> str:
    """
    Gets text from the decision of source in PDF. Avoids the first page if it is a cover page.
    :param options:
    :param source:
    :return:
    """
    r = requests.get(source.download_url)
    from pdfminer.high_level import extract_text_to_fp
    from pdfminer.layout import LAParams
    with PDFFile(source, options) as pdf, io.StringIO() as output_string, io.StringIO() as test_string:
        params = LAParams(line_margin=2)
        extract_text_to_fp(pdf, test_string, page_numbers=[0])
        first_page = test_string.getvalue()
        if len(first_page.split()) > 100:
            extract_text_to_fp(pdf, output_string, laparams=params)
        else:
            from pdfminer.pdfpage import PDFPage
            pages = len([page for page in PDFPage.get_pages(pdf)])
            extract_text_to_fp(pdf, output_string, page_numbers=[i for i in range(pages)[1:]], laparams=params)
        return output_string.getvalue()


def get_text_stream(item):
    from bs4 import BeautifulSoup
    soup = BeautifulSoup(requests.get(item.download_url).text, features='html5lib')
    rte = soup.find('div', class_='rte')
    text = rte.get_text()
    assert text, 'Download_text failed to get text from web page.'
    return text



#  MIT License Copyright (c) 2020. Houfu Ang

import logging
import os
import time

import click

from pdpc_decisions.classes import Options
from pdpc_decisions.corpus_text import create_corpus
from pdpc_decisions.download_file import download_files
from pdpc_decisions.save_file import save_scrape_results_to_csv
from pdpc_decisions.scraper import Scraper
from pdpc_decisions.scraper_extras import scraper_extras

logger = logging.getLogger(__name__)


@click.command()
@click.option('--csv', help='Filename for saving the items gathered by scraper as a csv file.',
              default='scrape_results.csv', type=click.Path(dir_okay=False), show_default=True)
@click.option('--download', help='Destination folder for downloads of all PDF/web pages of PDPC decisions',
              default='download/', type=click.Path(file_okay=False), show_default=True)
@click.option('--corpus', help='Destination folder for PDPC decisions converted to text files', default='corpus/',
              type=click.Path(file_okay=False), show_default=True)
@click.option('--root', '-r', help='Root directory for downloads and files', type=click.Path(file_okay=False),
              default=os.getcwd(), show_default=True)
@click.option('--extras/--no-extras', default=False, help='Add extra features to the data scraped', show_default=True)
@click.option('--extra-corpus/--no-extra-corpus', default=False, help='Enable experimental features for corpus',
              show_default=True)
@click.option('--verbose', '-v', default=False, help='Verbose output', show_default=True, is_flag=True)
@click.argument('action')
def pdpc_decision(csv, download, corpus, action, root, extras, extra_corpus, verbose):
    """
    Scripts to scrape all decisions of the Personal Data Protection Commission of Singapore.

    Accepts the following actions.

    "all"       Does all the actions (scraping the website, saving a csv, downloading all files and creating a corpus).

    "corpus"    After downloading all the decisions from the website, converts them into text files.

    "csv"      Save the items gathered by the scraper as a csv file.

    "files"     Downloads all the decisions from the PDPC website into a folder.
    """
    start_time = time.time()
    if verbose:
        logging.basicConfig(level='INFO')
    options = Options(csv_path=csv, download_folder=download, corpus_folder=corpus, action=action, root=root,
                      extras=extras, extra_corpus=extra_corpus)
    logger.info(f'Options: {options}')
    if options['root']:
        os.chdir(root)
    scrape_results = Scraper.scrape()
    if (action == 'all') or (action == 'files'):
        download_files(options, scrape_results)
    if (action == 'all') or (action == 'corpus'):
        create_corpus(options, scrape_results)
    if extras and ((action == 'all') or (action == 'csv')):
        scraper_extras(scrape_results, options)
    if (action == 'all') or (action == 'csv'):
        save_scrape_results_to_csv(options, scrape_results)
    diff = time.time() - start_time
    logger.info('Finished. This took {}s.'.format(diff))


if __name__ == '__main__':
    pdpc_decision()



#  MIT License Copyright (c) 2020. Houfu Ang

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pdpc-decisions',
    version='1.3.1',
    description='Tools to extract and compile enforcement '
                'decisions from the Singapore Personal Data Protection Commission',
    author='Ang Houfu ',
    author_email='houfu@outlook.sg',
    url='https://github.com/houfu/pdpc-decisions/',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['pdpc_decisions'],
    install_requires=['Click', 'selenium', 'beautifulsoup4', 'pdfminer.six', 'html5lib',
                      'requests', 'spacy'],
    classifiers=[
        'Development Status :: 5 - Production/Stable'
    ],
    python_requires='>=3.8',
    entry_points='''
    [console_scripts]
    pdpc_decisions=pdpc_decisions.pdpcdecision:pdpc_decision
''',
)



import io
import logging
import re
from dataclasses import dataclass
from datetime import datetime
from typing import TypedDict, List
from urllib.parse import urljoin

import bs4
import requests

logger = logging.getLogger(__name__)


class Options(TypedDict):
    csv_path: str
    download_folder: str
    corpus_folder: str
    action: str
    root: str
    extras: bool
    extra_corpus: bool


@dataclass
class PDPCDecisionItem:
    published_date: datetime.date
    respondent: str
    title: str
    summary: str
    download_url: str

    @classmethod
    def from_web_page(cls, decision: str):
        """
        Create a PDPCDecisionItem from a section in the PDPC Website's list of commission's decisions.
        :param decision: Link to PDPC's decision page.
        :return:
        """
        soup = bs4.BeautifulSoup(requests.get(decision).text, features='html5lib')
        article = soup.find('article')
        published_date = cls.get_published_date(article)
        respondent = cls.get_respondent(article)
        title = cls.get_title(article)
        summary = cls.get_summary(article)
        download_url = cls.get_url(article, decision)
        return cls(published_date, respondent, title, summary, download_url)

    @staticmethod
    def get_url(article: bs4.Tag, url: str) -> str:
        """Gets the URL for the text of the decision."""
        link = article.find('a')
        return urljoin(url, link['href'])

    @staticmethod
    def get_summary(article: bs4.Tag) -> str:
        """Gets the summary of a decision as provided by the PDPC."""
        paragraphs = article.find(class_='rte').find_all('p')
        result = ''
        for paragraph in paragraphs:
            if not paragraph.text == '':
                result += re.sub(r'\s+', ' ', paragraph.text)
                break
        return result

    @staticmethod
    def get_published_date(article: bs4.Tag) -> datetime.date:
        """Gets the date when the decision is published on the PDPC Website"""
        return datetime.strptime(article.find(class_='page-date').text, "%d %b %Y").date()

    @staticmethod
    def get_respondent(article: bs4.Tag) -> str:
        """Gets the name of the respondent in the decision from title of the decision."""
        text = re.split(r"\s+[bB]y|[Aa]gainst\s+", article.find('h2').text, re.I)[1].strip()
        return re.sub(r'\s+', ' ', text)

    @staticmethod
    def get_title(article: bs4.Tag) -> str:
        """Gets the title of the decision as provided by the PDPC"""
        return re.sub(r'\s+', ' ', article.find('h2').text)

    def __str__(self):
        return f"PDPCDecisionItem: {self.published_date} {self.respondent}"


class Paragraph(object):
    def __init__(self, text: str, paragraph_mark: str = ''):
        self._text = text
        self._paragraph_mark = paragraph_mark

    @property
    def text(self):
        """Get the text of this paragraph."""
        return self._text

    @property
    def paragraph_mark(self):
        """Get the mark (paragraph number) of this paragraph, if available."""
        return self._paragraph_mark

    def update_text(self, new_text: str):
        self._text = new_text

    def __str__(self):
        if self._paragraph_mark:
            return f"Paragraph: {self._paragraph_mark} {self._text}"
        else:
            return f"Paragraph: NA, {self._text}"


class CorpusDocument:
    def __init__(self, source: PDPCDecisionItem = None):
        """
        Base document representing a PDPC Decision.
        :param source: (Optional) The decision which this document represents.
        """
        self.source = source
        self.paragraphs: List[Paragraph] = []

    def get_text_as_paragraphs(self, add_paragraph_marks: bool = False) -> List[str]:
        """
        Convenience method to get the text of the corpus_text as a list of string representing paragraphs.
        :param add_paragraph_marks: Paragraphs marks (such as paragraph numbers) are added to the front
        of the paragraph.
        :return:
        """
        result = []
        for paragraph in self.paragraphs:
            if add_paragraph_marks:
                result.append(f"{paragraph.paragraph_mark} {paragraph.text}")
            else:
                result.append(f"{paragraph.text}")
        return result

    def get_text(self, add_paragraph_marks: bool = False) -> str:
        """
        Convenience method to get the text of the corpus_text as string.
        :param add_paragraph_marks: Paragraphs marks (such as paragraph numbers) are added to the front
        of the paragraph.
        :return:
        """
        return " ".join(self.get_text_as_paragraphs(add_paragraph_marks))

    def add_paragraph(self, text: str, mark: str = ''):
        self.paragraphs.append(Paragraph(text, mark))

    def __iter__(self):
        return iter(self.paragraphs)

    def __str__(self):
        if self.source:
            return f"CorpusDocument:{len(self.paragraphs)}, source: {self.source.respondent}, " \
                   f"{self.source.published_date}"
        else:
            return f"CorpusDocument:{len(self.paragraphs)}, source: None"


class PDFFile:
    def __init__(self, source: PDPCDecisionItem, options: Options = None, *args, **kwargs):
        """
        A context manager to facilitate opening of PDF files. If a local file is found, that file is
        returned instead of downloading from the source.
        :param source:
        :param options:
        :param args:
        :param kwargs:
        """
        if options and options['download_folder']:
            import os
            pdf_local_file_path = os.path.join(os.getcwd(), options['download_folder'],
                                               f"{source.published_date.strftime('%Y-%m-%d')} {source.respondent}.pdf")
            if os.path.isfile(pdf_local_file_path):
                logger.info('Using PDF file from local directory.')
                self.file_handler = open(pdf_local_file_path, 'rb')
                return
            if os.path.isfile(
                    os.path.join(os.getcwd(), options['download_folder'],
                                 f"{source.published_date.strftime('%Y-%m-%d')} {source.respondent} (1).pdf")):
                logger.warning(f'Returning a local file is not supported for secondary format. Source: {str(source)}')
        import requests
        r = requests.get(source.download_url)
        logger.info('Using PDF file from source.')
        self.file_handler = io.BytesIO(r.content)

    def __enter__(self):
        return self.file_handler

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file_handler.close()
        return False
